@@ -38,6 +38,132 @@ const char *grn_document_root = NULL;
 
 #define VAR GRN_PROC_GET_VAR_BY_OFFSET
 
+/**** query expander ****/
+
+static grn_rc
+substitute_query(grn_ctx *ctx, grn_obj *table, grn_obj *column,
+                 const char *key, size_t key_size, grn_obj *dest)
+{
+  grn_id id;
+  grn_rc rc = GRN_END_OF_DATA;
+  if ((id = grn_table_get(ctx, table, (const void *)key, (unsigned)key_size))) {
+    grn_obj_get_value(ctx, column, id, dest);
+    rc = GRN_SUCCESS;
+  }
+  return rc;
+}
+
+static grn_rc
+expand_query(grn_ctx *ctx, grn_obj *table, grn_obj *column, grn_expr_flags flags,
+             const char *str, unsigned str_len, grn_obj *dest)
+{
+  grn_obj buf;
+  unsigned int len;
+  const char *start, *cur = str, *str_end = str + (size_t)str_len;
+  GRN_TEXT_INIT(&buf, 0);
+  for (;;) {
+    while (cur < str_end && grn_isspace(cur, ctx->encoding)) {
+      if (!(len = grn_charlen(ctx, cur, str_end))) { goto exit; }
+      GRN_TEXT_PUT(ctx, dest, cur, len);
+      cur += len;
+    }
+    if (str_end <= cur) { break; }
+    switch (*cur) {
+    case '\0' :
+      goto exit;
+      break;
+    case GRN_QUERY_AND :
+    case GRN_QUERY_ADJ_INC :
+    case GRN_QUERY_ADJ_DEC :
+    case GRN_QUERY_ADJ_NEG :
+    case GRN_QUERY_BUT :
+    case GRN_QUERY_PARENL :
+    case GRN_QUERY_PARENR :
+    case GRN_QUERY_PREFIX :
+      GRN_TEXT_PUTC(ctx, dest, *cur);
+      cur++;
+      break;
+    case GRN_QUERY_QUOTEL :
+      GRN_BULK_REWIND(&buf);
+      for (start = cur++; cur < str_end; cur += len) {
+        if (!(len = grn_charlen(ctx, cur, str_end))) {
+          goto exit;
+        } else if (len == 1) {
+          if (*cur == GRN_QUERY_QUOTER) {
+            cur++;
+            break;
+          } else if (cur + 1 < str_end && *cur == GRN_QUERY_ESCAPE) {
+            cur++;
+            len = grn_charlen(ctx, cur, str_end);
+          }
+        }
+        GRN_TEXT_PUT(ctx, &buf, cur, len);
+      }
+      if (substitute_query(ctx, table, column, GRN_TEXT_VALUE(&buf), GRN_TEXT_LEN(&buf), dest)) {
+        GRN_TEXT_PUT(ctx, dest, start, cur - start);
+      }
+      break;
+    case 'O' :
+      if (cur + 2 <= str_end && cur[1] == 'R' &&
+          (cur + 2 == str_end || grn_isspace(cur + 2, ctx->encoding))) {
+        GRN_TEXT_PUT(ctx, dest, cur, 2);
+        cur += 2;
+        break;
+      }
+      /* fallthru */
+    default :
+      for (start = cur; cur < str_end; cur += len) {
+        if (!(len = grn_charlen(ctx, cur, str_end))) {
+          goto exit;
+        } else if (grn_isspace(cur, ctx->encoding)) {
+          break;
+        } else if (len == 1) {
+          if (*cur == GRN_QUERY_PARENL ||
+              *cur == GRN_QUERY_PARENR ||
+              *cur == GRN_QUERY_PREFIX) {
+            break;
+          } else if (flags & GRN_EXPR_ALLOW_COLUMN && *cur == GRN_QUERY_COLUMN) {
+            if (cur + 1 < str_end) {
+              switch (cur[1]) {
+              case '!' :
+              case '@' :
+              case '^' :
+              case '$' :
+                cur += 2;
+                break;
+              case '=' :
+                cur += (flags & GRN_EXPR_ALLOW_UPDATE) ? 2 : 1;
+                break;
+              case '<' :
+              case '>' :
+                cur += (cur + 2 < str_end && cur[2] == '=') ? 3 : 2;
+                break;
+              default :
+                cur += 1;
+                break;
+              }
+            } else {
+              cur += 1;
+            }
+            GRN_TEXT_PUT(ctx, dest, start, cur - start);
+            start = cur;
+            break;
+          }
+        }
+      }
+      if (start < cur) {
+        if (substitute_query(ctx, table, column, start, cur - start, dest)) {
+          GRN_TEXT_PUT(ctx, dest, start, cur - start);
+        }
+      }
+      break;
+    }
+  }
+exit :
+  GRN_OBJ_FIN(ctx, &buf);
+  return GRN_SUCCESS;
+}
+
 /**** procs ****/
 
 #define DEFAULT_LIMIT           10
@@ -60,7 +186,8 @@ grn_select(grn_ctx *ctx, const char *table, unsigned table_len,
            const char *drilldown_output_columns, unsigned drilldown_output_columns_len,
            int drilldown_offset, int drilldown_limit,
            const char *cache, unsigned cache_len,
-           const char *match_escalation_threshold, unsigned match_escalation_threshold_len)
+           const char *match_escalation_threshold, unsigned match_escalation_threshold_len,
+           const char *query_expansion, unsigned query_expansion_len)
 {
   uint32_t nkeys, nhits;
   uint16_t cacheable = 1, taintable = 0;
@@ -73,7 +200,7 @@ grn_select(grn_ctx *ctx, const char *table, unsigned table_len,
   uint32_t cache_key_size = table_len + 1 + match_columns_len + 1 + query_len + 1 +
     filter_len + 1 + scorer_len + 1 + sortby_len + 1 + output_columns_len + 1 +
     drilldown_len + 1 + drilldown_sortby_len + 1 + drilldown_output_columns_len +
-    match_escalation_threshold_len + 1 +
+    match_escalation_threshold_len + 1 + query_expansion_len + 1 +
     sizeof(grn_content_type) + sizeof(int) * 4;
   long long int threshold, original_threshold;
   if (cache_key_size <= GRN_TABLE_MAX_KEY_SIZE) {
@@ -99,8 +226,10 @@ grn_select(grn_ctx *ctx, const char *table, unsigned table_len,
     cp += drilldown_sortby_len; *cp++ = '\0';
     memcpy(cp, drilldown_output_columns, drilldown_output_columns_len);
     cp += drilldown_output_columns_len; *cp++ = '\0';
-    memcpy(cp, &match_escalation_threshold, match_escalation_threshold_len);
+    memcpy(cp, match_escalation_threshold, match_escalation_threshold_len);
     cp += match_escalation_threshold_len; *cp++ = '\0';
+    memcpy(cp, query_expansion, query_expansion_len);
+    cp += query_expansion_len; *cp++ = '\0';
     memcpy(cp, &output_type, sizeof(grn_content_type)); cp += sizeof(grn_content_type);
     memcpy(cp, &offset, sizeof(int)); cp += sizeof(int);
     memcpy(cp, &limit, sizeof(int)); cp += sizeof(int);
@@ -139,9 +268,36 @@ grn_select(grn_ctx *ctx, const char *table, unsigned table_len,
           }
         }
         if (query_len) {
+          grn_expr_flags flags;
+          grn_obj query_expansion_buf;
+          GRN_TEXT_INIT(&query_expansion_buf, 0);
+          flags = GRN_EXPR_SYNTAX_QUERY|GRN_EXPR_ALLOW_PRAGMA|GRN_EXPR_ALLOW_COLUMN;
+          if (query_expansion_len) {
+            grn_obj *query_expansion_column;
+            query_expansion_column = grn_ctx_get(ctx, query_expansion, query_expansion_len);
+            if (query_expansion_column) {
+              grn_obj *query_expansion_table;
+              query_expansion_table = grn_column_table(ctx, query_expansion_column);
+              if (query_expansion_table) {
+                expand_query(ctx, query_expansion_table, query_expansion_column, flags,
+                             query, query_len, &query_expansion_buf);
+                query = GRN_TEXT_VALUE(&query_expansion_buf);
+                query_len = GRN_TEXT_LEN(&query_expansion_buf);
+                grn_obj_unlink(ctx, query_expansion_table);
+              }
+              grn_obj_unlink(ctx, query_expansion_column);
+            } else {
+              ERR(GRN_INVALID_ARGUMENT,
+                  "nonexistent query expansion column: <%.*s>",
+                  query_expansion_len, query_expansion);
+              grn_obj_unlink(ctx, cond);
+              GRN_OBJ_FIN(ctx, &query_expansion_buf);
+              goto exit;
+            }
+          }
           grn_expr_parse(ctx, cond, query, query_len,
-                         match_columns_, GRN_OP_MATCH, GRN_OP_AND,
-                         GRN_EXPR_SYNTAX_QUERY|GRN_EXPR_ALLOW_PRAGMA|GRN_EXPR_ALLOW_COLUMN);
+                         match_columns_, GRN_OP_MATCH, GRN_OP_AND, flags);
+          GRN_OBJ_FIN(ctx, &query_expansion_buf);
           if (!ctx->rc && filter_len) {
             grn_expr_parse(ctx, cond, filter, filter_len,
                            match_columns_, GRN_OP_MATCH, GRN_OP_AND,
@@ -298,6 +454,7 @@ grn_select(grn_ctx *ctx, const char *table, unsigned table_len,
   } else {
     ERR(GRN_INVALID_ARGUMENT, "invalid table name: <%.*s>", table_len, table);
   }
+exit:
   if (match_escalation_threshold_len) {
     grn_ctx_set_match_escalation_threshold(ctx, original_threshold);
   }
@@ -345,7 +502,8 @@ proc_select(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
                  drilldown_output_columns, drilldown_output_columns_len,
                  drilldown_offset, drilldown_limit,
                  GRN_TEXT_VALUE(VAR(14)), GRN_TEXT_LEN(VAR(14)),
-                 GRN_TEXT_VALUE(VAR(15)), GRN_TEXT_LEN(VAR(15)))) {
+                 GRN_TEXT_VALUE(VAR(15)), GRN_TEXT_LEN(VAR(15)),
+                 GRN_TEXT_VALUE(VAR(16)), GRN_TEXT_LEN(VAR(16)))) {
   }
   return NULL;
 }
@@ -2343,7 +2501,7 @@ func_edit_distance(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_
 void
 grn_db_init_builtin_query(grn_ctx *ctx)
 {
-  grn_expr_var vars[17];
+  grn_expr_var vars[18];
 
   DEF_VAR(vars[0], "name");
   DEF_VAR(vars[1], "table");
@@ -2362,8 +2520,9 @@ grn_db_init_builtin_query(grn_ctx *ctx)
   DEF_VAR(vars[14], "drilldown_limit");
   DEF_VAR(vars[15], "cache");
   DEF_VAR(vars[16], "match_escalation_threshold");
-  DEF_COMMAND("define_selector", proc_define_selector, 17, vars);
-  DEF_COMMAND("select", proc_select, 16, vars + 1);
+  DEF_VAR(vars[17], "query_expansion");
+  DEF_COMMAND("define_selector", proc_define_selector, 18, vars);
+  DEF_COMMAND("select", proc_select, 17, vars + 1);
 
   DEF_VAR(vars[0], "values");
   DEF_VAR(vars[1], "table");
