@@ -51,7 +51,6 @@ grn_rc grn_ctx_close(grn_ctx *ctx);
 
 #define LISTEN_BACKLOG 756
 #define MIN_MAX_FDS 2048
-#define LOG_SPLIT_LINES 1000000
 #define MAX_THREADS 128 /* max 256 */
 
 typedef enum {
@@ -78,6 +77,7 @@ typedef struct {
   const char *log_base_path;
   FILE *log_file;
   uint32_t log_count;
+  grn_bool request_reopen_log_file;
 } thd_data;
 
 typedef struct {
@@ -89,13 +89,17 @@ typedef struct {
 
 #define CMD_BUF_SIZE 1024
 
+static thd_data threads[MAX_THREADS];
 static uint32_t default_max_threads = DEFAULT_MAX_THREADS;
+static uint32_t max_threads;
 static volatile sig_atomic_t loop = 1;
 static grn_obj *db;
+static uint32_t n_lines_per_log_file = 1000000;
 
 static int
 suggest_result(struct evbuffer *res_buf, const char *types, const char *query,
-               const char *target_name, int threshold, int limit,
+               const char *target_name, int frequency_threshold,
+               double conditional_probability_threshold, int limit,
                grn_obj *cmd_buf, grn_ctx *ctx)
 {
   if (target_name && types && query) {
@@ -106,8 +110,10 @@ suggest_result(struct evbuffer *res_buf, const char *types, const char *query,
     grn_text_urlenc(ctx, cmd_buf, types, strlen(types));
     GRN_TEXT_PUTS(ctx, cmd_buf, "&query=");
     grn_text_urlenc(ctx, cmd_buf, query, strlen(query));
-    GRN_TEXT_PUTS(ctx, cmd_buf, "&threshold=");
-    grn_text_itoa(ctx, cmd_buf, threshold);
+    GRN_TEXT_PUTS(ctx, cmd_buf, "&frequency_threshold=");
+    grn_text_itoa(ctx, cmd_buf, frequency_threshold);
+    GRN_TEXT_PUTS(ctx, cmd_buf, "&conditional_probability_threshold=");
+    grn_text_ftoa(ctx, cmd_buf, conditional_probability_threshold);
     GRN_TEXT_PUTS(ctx, cmd_buf, "&limit=");
     grn_text_itoa(ctx, cmd_buf, limit);
     {
@@ -132,12 +138,14 @@ log_send(struct evkeyvalq *output_headers, struct evbuffer *res_buf,
          thd_data *thd, struct evkeyvalq *get_args)
 {
   uint64_t millisec;
-  int threshold, limit;
+  int frequency_threshold, limit;
+  double conditional_probability_threshold;
   const char *callback, *types, *query, *client_id, *target_name,
              *learn_target_name;
 
   parse_keyval(get_args, &query, &types, &client_id, &target_name,
-               &learn_target_name, &callback, &millisec, &threshold, &limit);
+               &learn_target_name, &callback, &millisec, &frequency_threshold,
+               &conditional_probability_threshold, &limit);
 
   /* send data to learn client */
   if (thd->zmq_sock && millisec && client_id && query && learn_target_name) {
@@ -212,15 +220,17 @@ log_send(struct evkeyvalq *output_headers, struct evbuffer *res_buf,
       evbuffer_add(res_buf, callback, content_length);
       evbuffer_add(res_buf, "(", 1);
       content_length += suggest_result(res_buf, types, query, target_name,
-                                       threshold, limit,
-                                       &(thd->cmd_buf), thd->ctx) + 3;
+                                       frequency_threshold,
+                                       conditional_probability_threshold,
+                                       limit, &(thd->cmd_buf), thd->ctx) + 3;
       evbuffer_add(res_buf, ");", 2);
     } else {
       evhttp_add_header(output_headers,
                         "Content-Type", "application/json; charset=UTF-8");
       content_length = suggest_result(res_buf, types, query, target_name,
-                                      threshold, limit,
-                                      &(thd->cmd_buf), thd->ctx);
+                                      frequency_threshold,
+                                      conditional_probability_threshold,
+                                      limit, &(thd->cmd_buf), thd->ctx);
     }
     if (content_length >= 0) {
       char num_buf[16];
@@ -248,6 +258,14 @@ cleanup_httpd_thread(thd_data *thd) {
   event_base_free(thd->base);
 }
 
+static void
+close_log_file(thd_data *thread)
+{
+  fclose(thread->log_file);
+  thread->log_file = NULL;
+  thread->request_reopen_log_file = GRN_FALSE;
+}
+
 static void
 generic_handler(struct evhttp_request *req, void *arg)
 {
@@ -280,6 +298,9 @@ generic_handler(struct evhttp_request *req, void *arg)
     /* logging */
     {
       if (thd->log_base_path) {
+        if (thd->log_file && thd->request_reopen_log_file) {
+          close_log_file(thd);
+        }
         if (!thd->log_file) {
           time_t n;
           struct tm *t_st;
@@ -301,9 +322,10 @@ generic_handler(struct evhttp_request *req, void *arg)
         }
         if (thd->log_file) {
           fprintf(thd->log_file, "%s\n", req->uri);
-          if (++thd->log_count >= LOG_SPLIT_LINES) {
-            fclose(thd->log_file);
-            thd->log_file = NULL;
+          thd->log_count++;
+          if (n_lines_per_log_file > 0 &&
+              thd->log_count >= n_lines_per_log_file) {
+            close_log_file(thd);
           }
         }
       }
@@ -351,6 +373,16 @@ signal_handler(int sig)
   loop = 0;
 }
 
+static void
+signal_reopen_log_file(int sig)
+{
+  uint32_t i;
+
+  for (i = 0; i < max_threads; i++) {
+    threads[i].request_reopen_log_file = GRN_TRUE;
+  }
+}
+
 void
 timeout_handler(int fd, short events, void *arg) {
   thd_data *thd = arg;
@@ -524,55 +556,54 @@ serve_threads(int nthreads, int port, const char *db_path, void *zmq_ctx,
 {
   int nfd;
   uint32_t i;
-  thd_data thds[nthreads];
-
   if ((nfd = bind_socket(port)) < 0) {
     print_error("cannot bind socket. please check port number with netstat.");
     return -1;
   }
 
   for (i = 0; i < nthreads; i++) {
-    memset(&thds[i], 0, sizeof(thds[i]));
-    if (!(thds[i].base = event_init())) {
+    memset(&threads[i], 0, sizeof(threads[i]));
+    threads[i].request_reopen_log_file = GRN_FALSE;
+    if (!(threads[i].base = event_init())) {
       print_error("error in event_init() on thread %d.", i);
     } else {
-      if (!(thds[i].httpd = evhttp_new(thds[i].base))) {
+      if (!(threads[i].httpd = evhttp_new(threads[i].base))) {
         print_error("error in evhttp_new() on thread %d.", i);
       } else {
         int r;
-        if ((r = evhttp_accept_socket(thds[i].httpd, nfd))) {
+        if ((r = evhttp_accept_socket(threads[i].httpd, nfd))) {
           print_error("error in evhttp_accept_socket() on thread %d.", i);
         } else {
           if (send_endpoint) {
-            if (!(thds[i].zmq_sock = zmq_socket(zmq_ctx, ZMQ_PUB))) {
+            if (!(threads[i].zmq_sock = zmq_socket(zmq_ctx, ZMQ_PUB))) {
               print_error("cannot create zmq_socket.");
-            } else if (zmq_connect(thds[i].zmq_sock, send_endpoint)) {
+            } else if (zmq_connect(threads[i].zmq_sock, send_endpoint)) {
               print_error("cannot connect zmq_socket.");
-              zmq_close(thds[i].zmq_sock);
-              thds[i].zmq_sock = NULL;
+              zmq_close(threads[i].zmq_sock);
+              threads[i].zmq_sock = NULL;
             } else {
               uint64_t hwm = 1;
-              zmq_setsockopt(thds[i].zmq_sock, ZMQ_HWM, &hwm, sizeof(uint64_t));
+              zmq_setsockopt(threads[i].zmq_sock, ZMQ_HWM, &hwm, sizeof(uint64_t));
             }
           } else {
-            thds[i].zmq_sock = NULL;
+            threads[i].zmq_sock = NULL;
           }
-          if (!(thds[i].ctx = grn_ctx_open(0))) {
+          if (!(threads[i].ctx = grn_ctx_open(0))) {
             print_error("error in grn_ctx_open() on thread %d.", i);
-          } else if (grn_ctx_use(thds[i].ctx, db)) {
+          } else if (grn_ctx_use(threads[i].ctx, db)) {
             print_error("error in grn_db_open() on thread %d.", i);
           } else {
-            GRN_TEXT_INIT(&(thds[i].cmd_buf), 0);
-            thds[i].log_base_path = log_base_path;
-            thds[i].thread_id = i;
-            evhttp_set_gencb(thds[i].httpd, generic_handler, &thds[i]);
-            evhttp_set_timeout(thds[i].httpd, 10);
+            GRN_TEXT_INIT(&(threads[i].cmd_buf), 0);
+            threads[i].log_base_path = log_base_path;
+            threads[i].thread_id = i;
+            evhttp_set_gencb(threads[i].httpd, generic_handler, &threads[i]);
+            evhttp_set_timeout(threads[i].httpd, 10);
             {
               struct timeval tv = {1, 0};
-              evtimer_set(&(thds[i].pulse), timeout_handler, &thds[i]);
-              evtimer_add(&(thds[i].pulse), &tv);
+              evtimer_set(&(threads[i].pulse), timeout_handler, &threads[i]);
+              evtimer_add(&(threads[i].pulse), &tv);
             }
-            if ((r = pthread_create(&(thds[i].thd), NULL, dispatch, thds[i].base))) {
+            if ((r = pthread_create(&(threads[i].thd), NULL, dispatch, threads[i].base))) {
               print_error("error in pthread_create() on thread %d.", i);
             }
           }
@@ -598,10 +629,10 @@ serve_threads(int nthreads, int port, const char *db_path, void *zmq_ctx,
 
   /* join all httpd thread */
   for (i = 0; i < nthreads; i++) {
-    if (thds[i].thd) {
-      pthread_join(thds[i].thd, NULL);
+    if (threads[i].thd) {
+      pthread_join(threads[i].thd, NULL);
     }
-    cleanup_httpd_thread(&(thds[i]));
+    cleanup_httpd_thread(&(threads[i]));
   }
   return 0;
 }
@@ -654,10 +685,13 @@ usage(FILE *output)
     "  -r, --receive-endpoint <receive endpoint> : receive endpoint\n"
     "                                              (ex. tcp://example.com:1235)\n"
     "  -l, --log-base-path <path prefix>         : log path prefix\n"
+    "  --n-lines-per-log-file <lines number>     : number of lines in a log file\n"
+    "                                              use 0 for disabling this\n"
+    "                                              (default: %d)\n"
     "  -d, --daemon                              : daemonize\n"
     "  --disable-max-fd-check                    : disable max FD check on start\n"
     "  -h, --help                                : show this message\n",
-    DEFAULT_PORT, default_max_threads);
+    DEFAULT_PORT, default_max_threads, n_lines_per_log_file);
 }
 
 int
@@ -667,6 +701,7 @@ main(int argc, char **argv)
   const char *max_threads_string = NULL, *port_string = NULL;
   const char *address;
   const char *send_endpoint = NULL, *recv_endpoint = NULL, *log_base_path = NULL;
+  const char *n_lines_per_log_file_string = NULL;
   int n_processed_args, flags = RUN_MODE_ENABLE_MAX_FD_CHECK;
   run_mode mode = run_mode_none;
 
@@ -685,6 +720,7 @@ main(int argc, char **argv)
       {'s', "send-endpoint", NULL, 0, getopt_op_none},
       {'r', "receive-endpoint", NULL, 0, getopt_op_none},
       {'l', "log-base-path", NULL, 0, getopt_op_none},
+      {'\0', "n-lines-per-log-file", NULL, 0, getopt_op_none},
       {'d', "daemon", NULL, run_mode_daemon, getopt_op_update},
       {'\0', "disable-max-fd-check", NULL, RUN_MODE_ENABLE_MAX_FD_CHECK,
        getopt_op_off},
@@ -697,6 +733,7 @@ main(int argc, char **argv)
     opts[5].arg = &send_endpoint;
     opts[6].arg = &recv_endpoint;
     opts[7].arg = &log_base_path;
+    opts[8].arg = &n_lines_per_log_file_string;
 
     n_processed_args = grn_str_getopt(argc, argv, opts, &flags);
   }
@@ -714,7 +751,6 @@ main(int argc, char **argv)
   } else {
     grn_ctx ctx;
     void *zmq_ctx;
-    int max_threads;
 
     if (max_threads_string) {
       max_threads = atoi(max_threads_string);
@@ -743,6 +779,24 @@ main(int argc, char **argv)
       }
     }
 
+    if (n_lines_per_log_file_string) {
+      int64_t n_lines;
+      n_lines = grn_atoll(n_lines_per_log_file_string,
+                          n_lines_per_log_file_string + strlen(n_lines_per_log_file_string),
+                          NULL);
+      if (n_lines < 0) {
+        print_error("--n-lines-per-log-file must be >= 0: <%s>",
+                    n_lines_per_log_file_string);
+        return(EXIT_FAILURE);
+      }
+      if (n_lines > UINT32_MAX) {
+        print_error("--n-lines-per-log-file must be <= %ld: <%s>",
+                    UINT32_MAX, n_lines_per_log_file_string);
+        return(EXIT_FAILURE);
+      }
+      n_lines_per_log_file = (uint32_t)n_lines;
+    }
+
     if (mode == run_mode_daemon) {
       daemonize();
     }
@@ -754,6 +808,7 @@ main(int argc, char **argv)
         signal(SIGTERM, signal_handler);
         signal(SIGINT, signal_handler);
         signal(SIGQUIT, signal_handler);
+        signal(SIGUSR1, signal_reopen_log_file);
 
         serve_threads(max_threads, port_no, argv[n_processed_args], zmq_ctx,
                       send_endpoint, recv_endpoint, log_base_path);
