@@ -42,6 +42,11 @@ void test_invalid_table_name(void);
 void data_each(void);
 void test_each(gconstpointer data);
 void test_vector_reference_column(void);
+void test_invalid_int32_value(void);
+void data_valid_geo_point_valid(void);
+void test_valid_geo_point_value(gconstpointer data);
+void data_invalid_geo_point_valid(void);
+void test_invalid_geo_point_value(gconstpointer data);
 
 static gchar *tmp_directory;
 static const gchar *database_path;
@@ -330,7 +335,7 @@ test_index_geo_point(void)
     "[{\"_key\": \"たかね\", \"location\": \"128514964x502419287\"}]";
   const gchar *load_takane_with_empty_location =
     "load --table Shops\n"
-    "[{\"_key\": \"たかね\", \"location\": \"\"}]";
+    "[{\"_key\": \"たかね\", \"location\": null}]";
 
   assert_send_command("table_create Shops TABLE_HASH_KEY ShortText");
   assert_send_command("column_create Shops location COLUMN_SCALAR WGS84GeoPoint");
@@ -494,7 +499,7 @@ test_invalid_table_name(void)
     "and contains only 0-9, A-Z, a-z, #, - or _: <_Users>",
     "load --table _Users\n"
     "[\n"
-    "[\"_key\": \"alice\"]\n"
+    "{\"_key\": \"alice\"}\n"
     "]");
 }
 
@@ -580,3 +585,104 @@ test_vector_reference_column(void)
     send_command("select Users"));
 }
 
+void
+test_invalid_int32_value(void)
+{
+  assert_send_command("table_create Users TABLE_NO_KEY");
+  assert_send_command("column_create Users age COLUMN_SCALAR Int32");
+  grn_test_assert_send_command_error(
+    context,
+    GRN_INVALID_ARGUMENT,
+    "failed to cast to <Int32>: <\"invalid number!\">",
+    "load --table Users\n"
+    "[\n"
+    "{\"age\": \"invalid number!\"}\n"
+    "]\n");
+  send_command("]\n");
+  cut_assert_equal_string(
+    "[[[1],"
+     "[[\"_id\",\"UInt32\"],"
+      "[\"age\",\"Int32\"]],"
+     "[1,0]]]",
+    send_command("select Users"));
+}
+
+void
+data_valid_geo_point_value(void)
+{
+#define ADD_DATUM(label, location)                                      \
+  gcut_add_datum(label " (" location ")",                               \
+                 "location", G_TYPE_STRING, location,                   \
+                 NULL)
+
+  ADD_DATUM("too large latitude", "324000000x502419287");
+  ADD_DATUM("too small latitude", "-324000000x502419287");
+  ADD_DATUM("too large longitude", "128514964x648000000");
+  ADD_DATUM("too small longitude", "128514964x-648000000");
+
+#undef ADD_DATUM
+}
+
+void
+test_valid_geo_point_value(gconstpointer data)
+{
+  assert_send_command("table_create Shops TABLE_HASH_KEY ShortText");
+  assert_send_command("column_create Shops location COLUMN_SCALAR WGS84GeoPoint");
+  cut_assert_equal_string(
+    "1",
+    send_command(
+      cut_take_printf("load --table Shops\n"
+                      "[\n"
+                      "{\"_key\": \"たかね\", \"location\": \"%s\"}\n"
+                      "]\n",
+                      gcut_data_get_string(data, "location"))));
+  cut_assert_equal_string(
+    cut_take_printf("[[[1],"
+                    "[[\"_id\",\"UInt32\"],"
+                    "[\"_key\",\"ShortText\"],"
+                    "[\"location\",\"WGS84GeoPoint\"]],"
+                    "[1,\"たかね\",\"%s\"]]]",
+                    gcut_data_get_string(data, "location")),
+    send_command("select Shops"));
+}
+
+void
+data_invalid_geo_point_value(void)
+{
+#define ADD_DATUM(label, location)                                      \
+  gcut_add_datum(label " (" location ")",                               \
+                 "location", G_TYPE_STRING, location,                   \
+                 NULL)
+
+  ADD_DATUM("too large latitude", "324000001x502419287");
+  ADD_DATUM("too small latitude", "-324000001x502419287");
+  ADD_DATUM("too large longitude", "128514964x648000001");
+  ADD_DATUM("too small longitude", "128514964x-648000001");
+
+#undef ADD_DATUM
+}
+
+void
+test_invalid_geo_point_value(gconstpointer data)
+{
+  assert_send_command("table_create Shops TABLE_HASH_KEY ShortText");
+  assert_send_command("column_create Shops location COLUMN_SCALAR WGS84GeoPoint");
+  grn_test_assert_send_command_error(
+    context,
+    GRN_INVALID_ARGUMENT,
+    cut_take_printf("failed to cast to <WGS84GeoPoint>: <\"%s\">",
+                    gcut_data_get_string(data, "location")),
+    cut_take_printf("load --table Shops\n"
+                    "[\n"
+                    "{\"_key\": \"たかね\", \"location\": \"%s\"}\n"
+                    "]\n",
+                    gcut_data_get_string(data, "location")));
+  send_command("]\n");
+  cut_assert_equal_string(
+    "[[[1],"
+     "[[\"_id\",\"UInt32\"],"
+      "[\"_key\",\"ShortText\"],"
+      "[\"location\",\"WGS84GeoPoint\"]],"
+     "[1,\"たかね\",\"0x0\"]]]",
+    send_command("select Shops"));
+}
